# -*- coding: utf-8 -*-
import unittest
from homoglyphs import (
    Categories, Languages, Homoglyphs,
    STRATEGY_LOAD, STRATEGY_IGNORE, STRATEGY_REMOVE
)


CIRILLIC_ES = u'с'
CIRILLIC_HE = u'х'


class TestCommon(unittest.TestCase):
    def test_detect_category(self):
        self.assertEqual(Categories.detect('d'), 'LATIN')
        self.assertEqual(Categories.detect(u'Д'), 'CYRILLIC')
        self.assertEqual(Categories.detect('?'), 'COMMON')

    def test_detect_language(self):
        self.assertIn('en', Languages.detect('d'))
        self.assertIn('ru', Languages.detect(u'Д'))
        self.assertEqual(Languages.detect('?'), set())

    def test_get_alphabet_cat(self):
        alphabet = Categories.get_alphabet(['LATIN'])
        self.assertIn('s', alphabet)
        self.assertIn('S', alphabet)
        self.assertNotIn(u'ё', alphabet)
        self.assertGreater(len(alphabet), 50)
        self.assertLess(len(alphabet), 1400)

        alphabet = Categories.get_alphabet(['CYRILLIC'])
        self.assertIn(u'ё', alphabet)
        self.assertIn(u'Ё', alphabet)
        self.assertNotIn('s', alphabet)
        self.assertGreater(len(alphabet), 50)
        self.assertLess(len(alphabet), 500)

    def test_get_alphabet_lang(self):
        alphabet = Languages.get_alphabet({'en'})
        self.assertIn('s', alphabet)
        self.assertIn('S', alphabet)
        self.assertNotIn(u'ё', alphabet)
        self.assertEqual(len(alphabet), 26 * 2)

        alphabet = Languages.get_alphabet({'ru'})
        self.assertIn(u'ё', alphabet)
        self.assertIn(u'Ё', alphabet)
        self.assertNotIn('s', alphabet)
        self.assertEqual(len(alphabet), 33 * 2)

    def test_get_table(self):
        alphabet = Categories.get_alphabet(['LATIN'])
        table = Homoglyphs.get_table(alphabet)
        self.assertIn('s', table)
        self.assertNotIn(CIRILLIC_ES, table)

    def test_get_char_variants(self):
        variants = Homoglyphs(['LATIN'])._get_char_variants('s')
        self.assertIn('s', variants)
        self.assertIn(u'ｓ', variants)
        self.assertNotIn('S', variants)
        self.assertNotIn(CIRILLIC_ES, variants)

        variants = Homoglyphs(['LATIN'])._get_char_variants('c')
        self.assertIn('c', variants)
        self.assertNotIn(CIRILLIC_ES, variants)

        variants = Homoglyphs(['LATIN', 'CYRILLIC'])._get_char_variants('c')
        self.assertIn('c', variants)
        self.assertIn(CIRILLIC_ES, variants)

    def test_strategy(self):
        variants = Homoglyphs(['LATIN'], strategy=STRATEGY_IGNORE)._get_char_variants(u'ё')
        self.assertEqual(variants, [u'ё'])

        variants = Homoglyphs(['LATIN'], strategy=STRATEGY_LOAD)._get_char_variants(CIRILLIC_HE)
        self.assertGreater(len(variants), 1)
        self.assertIn('x', variants)
        self.assertIn(CIRILLIC_HE, variants)

        variants = Homoglyphs(['LATIN'], strategy=STRATEGY_REMOVE)._get_char_variants(CIRILLIC_HE)
        self.assertEqual(variants, [])

        for s in (STRATEGY_REMOVE, STRATEGY_IGNORE, STRATEGY_LOAD):
            variants = Homoglyphs(['LATIN'], strategy=s)._get_char_variants('d')
            self.assertIn('d', variants)

    def test_to_ascii(self):
        ss = Homoglyphs(strategy=STRATEGY_LOAD).to_ascii(CIRILLIC_HE)
        self.assertEqual(ss, ['x'])

        ss = Homoglyphs(strategy=STRATEGY_LOAD).to_ascii(CIRILLIC_HE)
        self.assertEqual(ss, ['x'])

        ss = Homoglyphs(strategy=STRATEGY_LOAD).to_ascii(CIRILLIC_HE + u'23.')
        self.assertEqual(ss, ['x23.'])

        ss = Homoglyphs(
            categories=('LATIN', 'COMMON', 'CYRILLIC'),
            ascii_strategy=STRATEGY_IGNORE,
        ).to_ascii(u'xхч2')
        self.assertEqual(ss, [])

        ss = Homoglyphs(
            categories=('LATIN', 'COMMON', 'CYRILLIC'),
            ascii_strategy=STRATEGY_REMOVE,
        ).to_ascii(u'xхч2')
        self.assertEqual(ss, ['xx2'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name='homoglyphs',
    version='1.2.2',

    author='orsinium',
    author_email='master_fess@mail.ru',

    description='Get homoglyphs for text, convert text to ASCII.',
    long_description=open('README.rst').read(),
    keywords='homoglyphs ascii utf8 text homoglyph similar letters',

    packages=['homoglyphs'],
    package_data={'': ['*.json']},
    requires=['python (>= 2.7)'],

    url='https://github.com/orsinium/homoglyphs',
    download_url='https://github.com/orsinium/homoglyphs/tarball/master',

    license='GNU Lesser General Public License v3.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
    ],
)



# -*- coding: utf-8 -*-
from collections import defaultdict
import json
from itertools import product
import os
import unicodedata

try:
    xrange
except NameError:
    # python3
    pass
else:
    # python2
    range = xrange  # noQA
    chr = unichr    # noQA


CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

# Actions if char not in alphabet
STRATEGY_LOAD = 1       # load category for this char
STRATEGY_IGNORE = 2     # add char to result
STRATEGY_REMOVE = 3     # remove char from result


class Categories(object):
    """
    Work with aliases from ISO 15924.
    https://en.wikipedia.org/wiki/ISO_15924#List_of_codes
    """
    fpath = os.path.join(CURRENT_DIR, 'categories.json')

    @classmethod
    def _get_ranges(cls, categories):
        """
        :return: iter: (start code, end code)
        :rtype: list
        """
        with open(cls.fpath) as f:
            data = json.load(f)
        indexes = []
        for category in categories:
            try:
                indexes.append(data['iso_15924_aliases'].index(category))
            except ValueError:
                raise ValueError('Invalid category: {}'.format(category))
        for point in data['code_points_ranges']:
            if point[2] in indexes:
                yield point[:2]

    @classmethod
    def get_alphabet(cls, categories):
        """
        :return: set of chars in alphabet by categories list
        :rtype: set
        """
        alphabet = set()
        for start, end in cls._get_ranges(categories):
            chars = (chr(code) for code in range(start, end))
            alphabet.update(set(chars))
        return set(alphabet)

    @classmethod
    def detect(cls, char):
        """
        :return: category
        :rtype: str
        """
        with open(cls.fpath) as f:
            data = json.load(f)

        # try detect category by unicodedata
        try:
            category = unicodedata.name(char).split()[0]
        except TypeError:
            # In Python2 unicodedata.name raise error for non-unicode chars
            pass
        else:
            if category in data['iso_15924_aliases']:
                return category

        # try detect category by ranges from JSON file.
        code = ord(char)
        for point in data['code_points_ranges']:
            if point[0] <= code <= point[1]:
                return data['iso_15924_aliases'][point[2]]


class Languages(object):
    fpath = os.path.join(CURRENT_DIR, 'languages.json')

    @classmethod
    def get_alphabet(cls, languages):
        """
        :return: set of chars in alphabet by languages list
        :rtype: set
        """
        with open(cls.fpath) as f:
            data = json.load(f)
        alphabet = set()
        for lang in languages:
            if lang not in data:
                raise ValueError('Invalid language code: {}'.format(lang))
            alphabet.update(data[lang])
        return alphabet

    @classmethod
    def detect(cls, char):
        """
        :return: set of languages which alphabet contains passed char.
        :rtype: set
        """
        with open(cls.fpath) as f:
            data = json.load(f)
        languages = set()
        for lang, alphabet in data.items():
            if char in alphabet:
                languages.add(lang)
        return languages


class Homoglyphs(object):
    def __init__(self, categories=None, languages=None, alphabet=None,
                 strategy=STRATEGY_IGNORE, ascii_strategy=STRATEGY_IGNORE):
        # strategies
        if strategy not in (STRATEGY_LOAD, STRATEGY_IGNORE, STRATEGY_REMOVE):
            raise ValueError('Invalid strategy')
        self.strategy = strategy
        self.ascii_strategy = ascii_strategy

        # Homoglyphs must be initialized by any alphabet for correct work
        if not categories and not languages and not alphabet:
            categories = ('LATIN', 'COMMON')

        # cats and langs
        self.categories = set(categories or [])
        self.languages = set(languages or [])

        # alphabet
        self.alphabet = set(alphabet or [])
        if self.categories:
            alphabet = Categories.get_alphabet(self.categories)
            self.alphabet.update(alphabet)
        if self.languages:
            alphabet = Languages.get_alphabet(self.languages)
            self.alphabet.update(alphabet)
        self.table = self.get_table(self.alphabet)

    @staticmethod
    def get_table(alphabet):
        table = defaultdict(list)
        with open(os.path.join(CURRENT_DIR, 'confusables.json')) as f:
            data = json.load(f)
        for char in alphabet:
            if char in data:
                for homoglyph in data[char]:
                    if homoglyph['c'] in alphabet:
                        table[char].append(homoglyph['c'])
        return table

    @staticmethod
    def uniq_and_sort(data):
        result = list(set(data))
        result.sort(key=lambda x: (-len(x), x))
        return result

    def _update_alphabet(self, char):
        # try detect languages
        langs = Languages.detect(char)
        if langs:
            self.languages.update(langs)
            alphabet = Languages.get_alphabet(langs)
            self.alphabet.update(alphabet)
        else:
            # try detect categories
            category = Categories.detect(char)
            if category is None:
                return False
            self.categories.add(category)
            alphabet = Categories.get_alphabet([category])
            self.alphabet.update(alphabet)
        # update table for new alphabet
        self.table = self.get_table(self.alphabet)
        return True

    def _get_char_variants(self, char):
        if char not in self.alphabet:
            if self.strategy == STRATEGY_LOAD:
                if not self._update_alphabet(char):
                    return []
            elif self.strategy == STRATEGY_IGNORE:
                return [char]
            elif self.strategy == STRATEGY_REMOVE:
                return []

        # find alternative chars for current char
        alt_chars = self.table.get(char, [])
        if alt_chars:
            # find alternative chars for alternative chars for current char
            alt_chars2 = [self.table.get(alt_char, []) for alt_char in alt_chars]
            # combine all alternatives
            alt_chars.extend(sum(alt_chars2, []))
        # add current char to alternatives
        alt_chars.append(char)

        if self.ascii_strategy == STRATEGY_REMOVE:
            alt_chars = [char for char in alt_chars if ord(char) < 256]

        # uniq, sort and return
        return self.uniq_and_sort(alt_chars)

    def _get_combinations(self, text):
        variations = []
        for char in text:
            alt_chars = self._get_char_variants(char)
            if alt_chars:
                variations.append(alt_chars)
        for variant in product(*variations):
            yield ''.join(variant)

    def get_combinations(self, text):
        return list(self._get_combinations(text))

    def _to_ascii(self, text):
        for variant in self._get_combinations(text):
            if max(map(ord, variant)) < 256:
                yield variant

    def to_ascii(self, text):
        return self.uniq_and_sort(self._to_ascii(text))
